%%%-------------------------------------------------------------------
%%% @author Dave Kuhlman
%%% @copyright (C) 2016, Dave Kuhlman
%%% @doc
%%%
%%% @end
%%% Created : 2016-12-19 14:06:45.650615
%%%-------------------------------------------------------------------
-module(db_server).

-behaviour(gen_server).

%% API
-export([start_link/0]).

%% gen_server callbacks
-export([init/1,
         handle_call/3,
         handle_cast/2,
         handle_info/2,
         terminate/2,
         code_change/3]).

%% db_server API
-export([
         get_rec/1,
         get_all_recs/0,
         delete_rec/1,
         create_rec/1,
         update_rec/2,
         rec_exists/1
        ]).

-define(SERVER, ?MODULE).

-record(state, {}).

%%%===================================================================
%%% API
%%%===================================================================

%%--------------------------------------------------------------------
%% @doc
%% Starts the server
%%
%% @spec start_link() -> {ok, Pid} | ignore | {error, Error}
%% @end
%%--------------------------------------------------------------------
start_link() ->
    gen_server:start_link({local, ?SERVER}, ?MODULE, [], []).

%%--------------------------------------------------------------------

get_rec(RecordId) ->
    %io:fwrite("(get_rec) RecordId: ~p~n", [RecordId]),
    gen_server:call(?SERVER, {get_rec, RecordId}).

get_all_recs() ->
    %io:fwrite("(get_all_recs)~n"),
    gen_server:call(?SERVER, {get_all_recs}).

delete_rec(RecordId) ->
    %io:fwrite("(delete_rec) RecordId: ~p~n", [RecordId]),
    gen_server:call(?SERVER, {delete_rec, RecordId}).

create_rec(Content) ->
    %io:fwrite("(create_rec) ~n", []),
    gen_server:call(?SERVER, {create_rec, Content}).

update_rec(RecordId, Content) ->
    %io:fwrite("(update_rec) RecordId: ~p  Content: ~p~n", [RecordId, Content]),
    gen_server:call(?SERVER, {update_rec, RecordId, Content}).

rec_exists(RecordId) ->
    %io:fwrite("(rec_exists) RecordId: ~p~n", [RecordId]),
    gen_server:call(?SERVER, {rec_exists, RecordId}).

%%%===================================================================
%%% gen_server callbacks
%%%===================================================================

%%--------------------------------------------------------------------
%% @private
%% @doc
%% Initializes the server
%%
%% @spec init(Args) -> {ok, State} |
%%                     {ok, State, Timeout} |
%%                     ignore |
%%                     {stop, Reason}
%% @end
%%--------------------------------------------------------------------
init([]) ->
    {ok, DataFileName} = application:get_env(data_file_name),
    %io:fwrite("(db_server:init) opening file: ~s~n", [DataFileName]),
    dets:open_file(data_db, [{file, DataFileName}, {type, set}]),
    {ok, #state{}}.

%%--------------------------------------------------------------------
%% @private
%% @doc
%% Handling call messages
%%
%% @spec handle_call(Request, From, State) ->
%%                                   {reply, Reply, State} |
%%                                   {reply, Reply, State, Timeout} |
%%                                   {noreply, State} |
%%                                   {noreply, State, Timeout} |
%%                                   {stop, Reason, Reply, State} |
%%                                   {stop, Reason, State}
%% @end
%%--------------------------------------------------------------------

handle_call({get_rec, Key}, _From, State) ->
    Records = dets:lookup(data_db, Key),
    %io:fwrite("(call get_rec) Records: ~p~n", [Records]),
    Reply = case Records of
        [{_RecordId, Data}] ->
            {ok, Data};
        [] ->
            {error, not_found};
        _ ->
            {error, too_many_records}
    end,
    {reply, Reply, State};

handle_call({get_all_recs}, _From, State) ->
    F = fun ({Id, Body}, Acc) -> Acc1 = [[Id, Body] | Acc], Acc1 end,
    Items = dets:foldl(F, [], data_db),
    Reply = {ok, Items},
    {reply, Reply, State};

handle_call({delete_rec, RecordId}, _From, State) ->
    Reply = case dets:lookup(data_db, RecordId) of
        [] ->
            {error, not_found};
        _ ->
            dets:delete(data_db, RecordId)
    end,
    {reply, Reply, State};

handle_call({create_rec, Content}, _From, State) ->
    %io:fwrite("(handle_call-create-rec) Content: ~p~n", [Content]),
    create_new_record(Content, State);

handle_call({update_rec, RecordId, NewContent}, _From, State) ->
    %io:fwrite("(handle_call-update-rec) RecordId: ~p  NewContent: ~p~n",
    %          [RecordId, NewContent]),
    Reply = case dets:delete(data_db, RecordId) of
        ok ->
            create_new_record(NewContent, State);
        {error, Reason} ->
            {reply, {error, Reason}, State}
    end,
    Reply;

handle_call({rec_exists, RecordId}, _From, State) ->
    Records = dets:lookup(data_db, RecordId),
    %io:fwrite("(call rec_exists) Records: ~p~n", [Records]),
    Reply = case Records of
        [{_RecordId, _Data}] ->
            ok;
        [] ->
            not_found;
        _ ->
            too_many_records
    end,
    {reply, Reply, State}.

%%--------------------------------------------------------------------
%% @private
%% @doc
%% Handling cast messages
%%
%% @spec handle_cast(Msg, State) -> {noreply, State} |
%%                                  {noreply, State, Timeout} |
%%                                  {stop, Reason, State}
%% @end
%%--------------------------------------------------------------------
handle_cast(_Msg, State) ->
        {noreply, State}.

%%--------------------------------------------------------------------
%% @private
%% @doc
%% Handling all non call/cast messages
%%
%% @spec handle_info(Info, State) -> {noreply, State} |
%%                                   {noreply, State, Timeout} |
%%                                   {stop, Reason, State}
%% @end
%%--------------------------------------------------------------------
handle_info(_Info, State) ->
        {noreply, State}.

%%--------------------------------------------------------------------
%% @private
%% @doc
%% This function is called by a gen_server when it is about to
%% terminate. It should be the opposite of Module:init/1 and do any
%% necessary cleaning up. When it returns, the gen_server terminates
%% with Reason. The return value is ignored.
%%
%% @spec terminate(Reason, State) -> void()
%% @end
%%--------------------------------------------------------------------
terminate(_Reason, _State) ->
    dets:close(data_db),
    ok.

%%--------------------------------------------------------------------
%% @private
%% @doc
%% Convert process state when code is changed
%%
%% @spec code_change(OldVsn, State, Extra) -> {ok, NewState}
%% @end
%%--------------------------------------------------------------------
code_change(_OldVsn, State, _Extra) ->
        {ok, State}.

%%%===================================================================
%%% Internal functions
%%%===================================================================

generate_id(Content) ->
    sha3:hexhash(256, Content).

create_new_record(Content, State) ->
    RecordId = generate_id(Content),
    ok = dets:insert(data_db, {RecordId, Content}),
    ok = dets:sync(data_db),
    Reply = {ok, RecordId},
    {reply, Reply, State}.
